'''This module evaluates an arithmetic expression input by the user, in
a supplied base. It uses the `Num` class of the `radix` module.'''

import re
from .radix import Num

SYMBOLS = ['+', '-', '/', '//', '*', '**', '%', '(', ')']


def expr(s, base=10, show=False):
    '''Evaluate the expression given by the user in the given base.

    Parameters -
    s - Expression string to be evaluated.
    base - The integer base of the numbers (should be between 2 and 36)
    show - Optionally print the expression that is input to `eval`.

    Note that using `eval` is safe here because only allowed operators
    are used and any other input is cast to the `Num` class. `eval` is
    not being used directly on the user input.
    '''
    tokens = tokenize(str(s))
    formatted_input = format_user_input(tokens, int(base))
    if show:
        print(formatted_input)
    return eval(formatted_input, {'Num': Num})


def tokenize(s):
    '''Get a list of operators and numbers, including floating point.'''

    return re.findall(r'(\w*[.]?\w+|//|\*\*|[()+*\-/])', s)


def format_user_input(tokens, base):
    '''Get a string formatted with allowed symbols and other input
    cast to the `Num` class.
    '''

    formatted_input = ''
    for token in tokens:
        if token in SYMBOLS:
            formatted_input += token
        else:
            formatted_token = f'Num(\'{token}\', {base})'
            formatted_input += formatted_token

    return formatted_input



'''This module represents signed binary numbers in 2's complement
form. It also enables 2's complement addition and subtraction.'''

from contextlib import contextmanager

from radix.radix import Num
from radix.complement import dim_radix_compl, radix_compl


@contextmanager
def reset_number(n: Num):

    value = n.value
    base = n.base
    yield
    n.value = value
    n.base = base


class Bin(Num):
    '''Signed binary numbers.

    Find representations of a number in binary 2's complement,
    1's complement and sign magnitude forms. Perform 2's complement
    addition and subtraction.
    '''

    def __init__(self, *args, twos_compl=None, **kwargs):
        '''Initialize binary number.

        If `twos_compl` parameter is passed, then only that is set on
        the instance. All other initialization is skipped. This is
        useful while adding and subtracting multiple binary numbers.
        '''
        if twos_compl is None:
            super().__init__(*args, **kwargs)

        self._2_compl = twos_compl or self.twos_compl()

    def __repr__(self):
        return self._2_compl

    def _process_base2(self):
        '''Create a base 2 number whose value has no sign and starts
        with 0.
        '''

        with reset_number(self):
            value = self.to(2).value.lstrip('+-')
        if value[0] in ('1', '.'):
            value = '0' + value
        n = Num(value, 2)
        return n

    def twos_compl(self):
        '''Find the 2's complement.'''

        n = self._process_base2()
        if self.sign() == -1:
            res = radix_compl(n)
            twos_compl_str = res[1:] if res.startswith('11') else res
        else:
            twos_compl_str = n.value

        return twos_compl_str

    def ones_compl(self):
        '''Find the 1's complement.'''

        n = self._process_base2()
        if self.sign() == -1:
            res = dim_radix_compl(n)
            ones_compl_str = res[1:] if res.startswith('11') else res
        else:
            ones_compl_str = n.value

        return ones_compl_str

    def sign_mag(self):
        '''Find the sign magnitude representation.'''

        with reset_number(self):
            value = self.to(2).value.lstrip('+-')
        if self.sign() == -1:
            return '1' + value
        elif value[0] in ('1', '.'):
            return '0' + value
        else:
            return value

    def format(self, size=8, blanks_every=None):
        '''Format the 2's complement string to be of a given size, with
        blanks inserted at intervals.
        '''

        value = self._2_compl
        fillchar = value[0]
        value = value.rjust(size, fillchar)

        if blanks_every:
            segments = []
            for i in range(0, len(value), blanks_every):
                segments.append(value[i:i + blanks_every])
                segments.append(' ')
            segments = segments[:-1]
            value = ''.join(segments)

        return value

    def sign(self):
        '''Sign of the binary number input to `Bin` class.'''

        if self.value.startswith('-'):
            return -1
        return 1

    def __add__(self, other):
        '''2's complement addition.'''

        a = self._2_compl
        b = other._2_compl
        twos_compl = self._add(a, b)
        return self.__class__(twos_compl=twos_compl)

    def _add(self, a, b):
        '''Perform addition of 2 binary numbers.'''

        # Make numbers equal in length
        if len(a) < len(b):
            a = a.rjust(len(b), a[0])
        else:
            b = b.rjust(len(a), b[0])

        # To track overflow
        if a[0] == b[0] == '0':
            flag = 0
        elif a[0] == b[0] == '1':
            flag = 1
        else:
            flag = -1

        a = a[::-1]
        b = b[::-1]
        carry = '0'
        r = ''  # intermediate result

        # Binary addition
        for i, j in zip(a, b):
            if i == '0' and j == '0':
                r += carry
                carry = '0'
            elif (i == '0' and j == '1') or (i == '1' and j == '0'):
                if carry == '1':
                    r += '0'
                else:
                    r += '1'
            else:
                if carry == '1':
                    r += '1'
                else:
                    r += '0'
                    carry = '1'
        r = r[::-1]
        # overflow
        if r[0] == '1' and flag == 0:
            result = '0' + r
        elif r[0] == '0' and flag == 1:
            result = '1' + r
        else:
            result = r

        return result

    def __sub__(self, other):
        '''2's complement subtraction.

        Add minuend to 2's complement of the subtrahend.
        '''

        a = self._2_compl
        b = (-other)._2_compl
        twos_compl = self._add(a, b)

        return self.__class__(twos_compl=twos_compl)

    @classmethod
    def from_Num(cls, instance: Num):

        # Prevent recalculation of base 10 value when Bin instance is
        # created.
        with reset_number(instance):
            instance.to(10)
            inst = cls(instance.value, 10)

        inst.value = instance.value
        inst.base = instance.base

        return inst

    def __getattr__(self, name):

        if name in ('value', 'base', 'base10_value'):
            raise NotImplementedError('Use `Num` class and convert the result '
                                      'to the `Bin` class with the `from_Num` '
                                      'classmethod.')

        raise AttributeError(f"'Bin' object has no attribute {name!r}")



'''This module converts numbers - both integer and floating-point -
from one base to another. It also enables arithmetic operations in a
given base.
'''
from string import digits, ascii_uppercase
import math

# 10 digits + 26 letters = 36 - upto base 36 is supported
DIGITS = digits + ascii_uppercase


class Num:
    '''Instantiate this class to represent a value in a certain base.
    Parameters -
    value - Can be a string or int or float in the given base.
    base - The base of the given value (should be between 2 and 36).
    '''

    def __init__(self, value, base=10):
        self.value = str(value).upper()
        self.base = int(base)
        if not (2 <= base <= 36):
            raise ValueError('Base must be an int between 2 and 36.')
        self.validate()
        if self.base == 10:
            self.base10_value = self._get_numeric_value(value)
        else:
            self.base10_value = self._to_base10()

    def _get_numeric_value(self, value):
        '''Convert possible string representation to number.'''

        if isinstance(value, str):
            try:
                return int(value)
            except ValueError:
                return float(value)

        return value

    def __repr__(self):
        return self.value

    def validate(self):
        '''Check if the user input is a valid number in the specified
        base.'''
        valid_digits = DIGITS[:self.base]
        for digit in self.value:
            if (digit not in ('.', '+', '-') and
                    not (self.base == 10 and digit in ('e', 'E'))):
                if digit not in valid_digits:
                    raise ValueError(f'{self.value} is an invalid base'
                                     f' {self.base} number.')

    def to(self, base, prec=10):
        '''Convert a number in a certain base to a new base, and return
        the instance. The stored base 10 value is used in the process.

        Parameters -
        base - The target base
        prec - The number of decimal places in the result if a float is
               being converted.
        '''
        if base == self.base:
            return self

        if base == 10:
            self.value = str(self.base10_value)
            self.base = 10
            return self

        frac_part, int_part = math.modf(self.base10_value)

        sign = '-' if int_part < 0 else ''
        int_part = abs(int(int_part))
        int_val = ''
        while int_part:
            int_part, rem = divmod(int_part, base)
            int_val = DIGITS[rem] + int_val
        if not int_val:
            int_val = '0'

        frac_part = abs(frac_part)
        frac_val = ''
        while frac_part and len(frac_val) < prec:
            val = frac_part * base
            frac_val += DIGITS[int(val)]
            frac_part = val - int(val)

        if frac_val or isinstance(self.base10_value, float):
            result = sign + int_val + '.' + (frac_val or '0')
        else:
            result = sign + int_val

        self.value = result
        self.base = base
        return self

    def _to_base10(self):
        '''Calculate the base 10 value of the stored number.'''

        int_part, *rest = self.value.split('.')
        frac_part = rest[0] if rest else ''

        int_value = 0
        sign = 1
        if int_part:
            c = int_part[0]
            if c in ('+', '-'):
                sign = -1 if c == '-' else 1
                int_part = int_part[1:]

        for index, digit in enumerate(int_part[::-1]):
            int_value += DIGITS.index(digit) * self.base ** index
        int_value *= sign

        frac_value = 0
        for index, digit in enumerate(frac_part, 1):
            frac_value += DIGITS.index(digit) * self.base ** -index
        frac_value *= sign

        if frac_value:
            result = int_value + frac_value
        else:
            result = int_value

        return result

    def __eq__(self, other):

        a, b = self._get_base10_values(other)
        if a == b:
            return True
        return False

    def __add__(self, other):

        self._check_bases(other)
        a, b = self._get_base10_values(other)
        result = a + b
        return self.__class__(result, 10).to(self.base)

    def __sub__(self, other):

        self._check_bases(other)
        a, b = self._get_base10_values(other)
        result = a - b
        return self.__class__(result, 10).to(self.base)

    def __mul__(self, other):

        self._check_bases(other)
        a, b = self._get_base10_values(other)
        result = a * b
        return self.__class__(result, 10).to(self.base)

    def __truediv__(self, other):

        self._check_bases(other)
        a, b = self._get_base10_values(other)
        result = a / b
        return self.__class__(result, 10).to(self.base)

    def __floordiv__(self, other):

        self._check_bases(other)
        a, b = self._get_base10_values(other)
        result = a // b
        return self.__class__(result, 10).to(self.base)

    def __mod__(self, other):

        self._check_bases(other)
        a, b = self._get_base10_values(other)
        result = a % b
        return self.__class__(result, 10).to(self.base)

    def __divmod__(self, other):

        self._check_bases(other)
        a, b = self._get_base10_values(other)
        q, r = divmod(a, b)
        return (self.__class__(q, 10).to(self.base),
                self.__class__(r, 10).to(self.base))

    def __pow__(self, other):

        self._check_bases(other)
        a, b = self._get_base10_values(other)
        result = a ** b
        return self.__class__(result, 10).to(self.base)

    def __neg__(self):

        a, _ = self._get_base10_values(self.__class__(-1))
        result = -1 * a
        return self.__class__(result, 10).to(self.base)

    def __pos__(self):

        return self

    def _get_base10_values(self, other):
        '''Return the base10 values.'''

        return self.base10_value, other.base10_value

    def _check_bases(self, other):
        '''Computations are possible when the operands have the same
        base.
        '''
        if self.base != other.base:
            raise TypeError(f'{self} and {other} don\'t have the same base.')



'''This module finds the r and (r-1) complements of a number in base
r.
'''

from radix.radix import Num, DIGITS


def prepare_value(n):

    value = n.value.lstrip('+-')

    adjusted_len = len(value)
    radix_pos = value.find('.')
    value = value.replace('.', '')

    return value, radix_pos, adjusted_len


def insert_radix_point(val, pos, length):

    if pos > -1:
        lst = list(val)
        lst.insert(pos, '.')
        r = ''.join(lst)
    else:
        r = val

    if len(r) < length:
        result = r.zfill(length)
    else:
        result = r

    return result


def dim_radix_compl(n):
    '''(r-1) or the diminished radix complement.'''

    value, radix_pos, adjusted_len = prepare_value(n)
    minuend = DIGITS[n.base - 1] * len(value)
    subtrahend = Num(value, n.base)
    r = Num(minuend, n.base) - subtrahend

    result = insert_radix_point(r.value, radix_pos, adjusted_len)

    return result


def radix_compl(n):
    '''r or the radix complement.'''

    if n == Num(0, n.base):
        return '0'

    r_minus_1_compl = dim_radix_compl(n)
    adjusted_len = len(r_minus_1_compl)
    radix_pos = r_minus_1_compl.find('.')
    value = r_minus_1_compl.replace('.', '')

    r = Num(value, n.base) + Num(1, n.base)
    result = insert_radix_point(r.value, radix_pos, adjusted_len)

    return result



import setuptools

with open('README.md', 'r') as fh:
    long_description = fh.read()

setuptools.setup(
    name='radix-ops',
    version='0.3.3',
    author='Debashish Palit',
    author_email='dpalit17@outlook.com',
    description='Convert between radices and perform arithmetic.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/deb17/radix',
    packages=['radix'],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.6'
)
